@@ -22,6 +22,13 @@
 
 /*
  * utility for getting accessible application UI info
+ *
+ * REFACTOR THOUGHTS
+ * The original plan was that gok-spy would be purely a library-like set of
+ * utility methods. Time has passed... hackery has happened...
+ * gok-spy contains some calls to gok-keyboard, and gok_main, and so is 
+ * starting to act a bit like a controller... this wasn't the original plan 
+ * but might actually make some sense.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -42,6 +49,9 @@
 #include "gok-word-complete.h"
 #include "gok-keyboard.h"
 
+/* gecko uses this utf8 embed */
+#define GECKO_EMBEDDED_OBJECT_CHAR 0xfffc
+
 /*
  * NOTE: we can tweak the efficiency/accuracy of the
  * gui search by adjusting using the spy/gui_search_depth and 
@@ -60,6 +70,7 @@
 
 /*
  * private prototypes
+1;5B
  */
 static void gok_spy_focus_listener (const AccessibleEvent *event,
                                     void *user_data);
@@ -90,7 +101,7 @@ static void gok_spy_object_state_listener (const AccessibleEvent *event,
                                                 void *user_data);
 
 static gboolean gok_spy_is_visible (Accessible *accessible);
-		
+
 static void gok_spy_free_nodes (GSList *nodelist);										
 static void gok_spy_resolve_namesakes (GSList *nodelist);
 
@@ -586,8 +597,7 @@ gok_spy_node_match (AccessibleNode *node, GokSpySearchType type)
 		return node->flags.data.is_ui;
 		break;
 	case GOK_SPY_SEARCH_EDITABLE_TEXT:
-		if (Accessible_isEditableText (node->paccessible))
-			return TRUE;
+		return gok_spy_is_editable (node->paccessible);
 		break;
 	case GOK_SPY_SEARCH_CHILDREN:
 	case GOK_SPY_SEARCH_COMBO:
@@ -643,14 +653,13 @@ gok_spy_get_editable (Accessible *search_root)
 	int child_count = Accessible_getChildCount (search_root);
 	while (i < child_count) 
 	{
-	    AccessibleRole role = SPI_ROLE_INVALID;
 	    child = Accessible_getChildAtIndex (search_root, i);
 	    if (child) 
 	    {
 		gok_spy_accessible_implicit_ref (child);
-		if (Accessible_isEditableText (child)) 
+		if (gok_spy_is_editable (child))
 		{
-		    return child;
+			return child;
 		}
 		else 
 		{
@@ -1035,6 +1044,73 @@ Accessible* gok_spy_get_accessibleWithText()
 	return m_ptheAccessibleWithText;
 }
 
+void _gok_spy_handle_window_activate(EventNode* en){
+	/* reset word completion, if it's being used */
+	gok_wordcomplete_reset (gok_wordcomplete_get_default ());
+	if (en->event->source != m_ptheWindowAccessible) {
+		gok_spy_accessible_unref (m_ptheWindowAccessible);
+		m_ptheWindowAccessible = en->event->source;
+		gok_spy_accessible_ref (m_ptheWindowAccessible);
+		if (m_ptheWindowChangeListener) {
+			((AccessibleChangeListenerCB)(m_ptheWindowChangeListener))(m_ptheWindowAccessible);
+		}
+	}
+}
+
+/* returned accessible needs unref */
+Accessible* _gok_spy_ancestor_with_role (Accessible* pAccessible, AccessibleRole role)
+{
+	Accessible* child = pAccessible;
+	Accessible* parent = NULL;
+	Accessible* ancestor = NULL;
+	
+	gok_spy_accessible_ref(child); /* need extra ref on passed in accessible */
+	gint generations = 0;
+	while (parent = Accessible_getParent(child)) {
+			gok_spy_accessible_unref(child);
+			if (Accessible_getRole(parent) == role) {
+				ancestor = parent;
+				break;
+			}
+			else if (generations++ > 25) {
+				break; /* avoid cyclic relations */
+			}
+			child = parent;
+	}
+	return ancestor;
+}
+
+/* for gecko: */
+Accessible* m_pLastBrowserPane = NULL;
+
+gboolean _gok_spy_handle_gecko_browse_tab_change (Accessible* source)
+{
+	/* gecko/ff3 tab pane handling:
+	gecko's tabs do not have the doc frames as children
+	so we need to capture focus in the containers and then
+	impose gok tab panel treatment.
+	*/
+	Accessible* browserpane = NULL;
+	GokKeyboard *current_kbd;
+	/* TODO look for attribute tag xul:tabbox instead a role? */
+	browserpane = (Accessible_getRole(source) == SPI_ROLE_DOCUMENT_FRAME)?
+	 source : _gok_spy_ancestor_with_role(source, SPI_ROLE_DOCUMENT_FRAME);
+	if (browserpane && browserpane != m_pLastBrowserPane){
+		if (m_pLastBrowserPane){Accessible_unref(m_pLastBrowserPane);}
+		m_pLastBrowserPane = browserpane;
+		/* this is one place we might need (in the future) to 
+		manually update gok_spy_update_component_list, passing in perhaps
+		the document frame or even the application root; but for now
+		it seems that all we need to do is tweak the main engine to redisplay
+		the keyboard; the component list will already be updated */
+		gok_main_display_scan_reset();
+/*		current_kbd = gok_main_get_current_keyboard ();
+		gok_main_ds (current_kbd);*/
+		return TRUE;
+	}
+	return FALSE;
+}
+
 /**
 * gok_spy_check_queues
 *
@@ -1073,14 +1149,15 @@ gboolean gok_spy_check_queues(void)
 			}
 		}
 #endif
-
 		switch (en->type)
 		{
 		case GOKSPY_FOCUS_EVENT:
 			/* TODO: remove obsolete focus events from queue since process
 			   focus is expensive. */
 			gok_log("GOKSPY_FOCUS_EVENT");
-			gok_spy_process_focus (en->event->source);
+			if (!_gok_spy_handle_gecko_browse_tab_change(en->event->source)){
+				gok_spy_process_focus (en->event->source);
+			}
 			break;
 		case GOKSPY_STATE_EVENT:
 			gok_log("GOKSPY_STATE_EVENT");
@@ -1090,16 +1167,7 @@ gboolean gok_spy_check_queues(void)
 			break;
 		case GOKSPY_WINDOW_ACTIVATE_EVENT:
 			gok_log("GOKSPY_WINDOW_ACTIVATE_EVENT");
-			/* reset word completion, if it's being used */
-			gok_wordcomplete_reset (gok_wordcomplete_get_default ());
-			if (en->event->source != m_ptheWindowAccessible) {
-				gok_spy_accessible_unref (m_ptheWindowAccessible);
-				m_ptheWindowAccessible = en->event->source;
-				gok_spy_accessible_ref (m_ptheWindowAccessible);
-				if (m_ptheWindowChangeListener) {
-					((AccessibleChangeListenerCB)(m_ptheWindowChangeListener))(m_ptheWindowAccessible);
-				}
-			}
+			_gok_spy_handle_window_activate(en);
 			break;
 		case GOKSPY_DEFUNCT_EVENT:
 			gok_log("GOKSPY_DEFUNCT_EVENT");
@@ -1169,18 +1237,20 @@ gboolean gok_spy_check_window(AccessibleRole role)
 {
 	/* TODO - improve efficiency here? Also, roles get added and we need
 	   to maintain this...  maybe we need to go about this differently */
-	if ((role == SPI_ROLE_WINDOW) ||
-	    (role == SPI_ROLE_DIALOG) ||
-	    (role == SPI_ROLE_FILE_CHOOSER) ||
-	    (role == SPI_ROLE_FRAME) ||	
-	    (role == SPI_ROLE_DESKTOP_FRAME) ||
-		(role == SPI_ROLE_FONT_CHOOSER)	||
-		(role == SPI_ROLE_COLOR_CHOOSER) ||
-		(role == SPI_ROLE_APPLICATION) ||
-		(role == SPI_ROLE_ALERT)
-		)
-	{
-		return TRUE;
+	switch (role) {
+		case SPI_ROLE_WINDOW:
+		case SPI_ROLE_DIALOG:
+		case SPI_ROLE_FILE_CHOOSER:
+		case SPI_ROLE_FRAME:
+		case SPI_ROLE_DESKTOP_FRAME:
+		case SPI_ROLE_FONT_CHOOSER:
+		case SPI_ROLE_COLOR_CHOOSER:
+		case SPI_ROLE_APPLICATION:
+		case SPI_ROLE_ALERT:
+		/* needs more testing case SPI_ROLE_HTML_CONTAINER:*/
+			return TRUE;
+		default:
+			break;
 	}
 	return FALSE;
 }
@@ -1201,42 +1271,30 @@ gboolean gok_spy_worth_searching (Accessible* accessible)
 	AccessibleStateSet* ass = NULL;
 
 	gok_log_enter();
-	
-	switch (Accessible_getRole (accessible)) 
-	{
-	case SPI_ROLE_MENU:
-	case SPI_ROLE_MENU_BAR:
-	case SPI_ROLE_MENU_ITEM:
-	case SPI_ROLE_CHECK_MENU_ITEM:
-	case SPI_ROLE_RADIO_MENU_ITEM:
-		bmenu = TRUE;
-		break;
-	default:
-		bmenu = FALSE;
-	}
-	ass = Accessible_getStateSet(accessible);
+
+	bmenu = gok_spy_is_menu_role(Accessible_getRole(accessible));	
+	/* no longer consider SPI_ROLE_MENU_BAR -- to work better with ff3*/
+
 	if (ass != NULL)
 	{
 		/* state heuristic: 
-		 * - if this thing manages its own children, then it 
+		 * - TODO: if this thing manages its own children, then it 
 		 * is probably not worth traversing.
 		 * - if it is a menu-like thing then it doesn't need to be "showing"
 		 */
 		if (!bmenu && !AccessibleStateSet_contains( ass, SPI_STATE_SHOWING ))
 		{
-			gchar *name = Accessible_getName (accessible);
-			gchar *role_name = Accessible_getRoleName (accessible);
-			gok_log ("%s %s is not worth searching", role_name, name);
 			worthiness = FALSE;
 		}
 		/* TODO: determine whether we can safely bypass manages-descendants objects */
 		/* it's not safe to exhaustviely search them */
-		if (AccessibleStateSet_contains( ass, SPI_STATE_MANAGES_DESCENDANTS ))
+		/* if (AccessibleStateSet_contains( ass, SPI_STATE_MANAGES_DESCENDANTS ))
 		{
 			gchar *name = Accessible_getName (accessible);
 			gchar *role_name = Accessible_getRoleName (accessible);
 			gok_log_x ("Manages-descendants state found on %s [%s]", name, role_name);
 		}
+		*/
 		AccessibleStateSet_unref(ass);
 	}
 	gok_log_leave();
@@ -1257,6 +1315,19 @@ gok_spy_is_visible (Accessible *accessible)
 	return retval;
 }
 
+gboolean
+gok_spy_is_editable (Accessible *accessible)
+{
+	gboolean retval = FALSE;
+	if (accessible && Accessible_isEditableText(accessible)) {
+		AccessibleStateSet *states = Accessible_getStateSet (accessible);
+		if (AccessibleStateSet_contains (states, SPI_STATE_EDITABLE))
+			retval = TRUE;
+		AccessibleStateSet_unref (states);
+	}
+	return retval;
+}
+
 gboolean
 gok_spy_is_menu_role (AccessibleRole role)
 {
@@ -1293,18 +1364,19 @@ gok_spy_is_ui (Accessible *accessible, Accessible *parent, AccessibleRole role)
 	case SPI_ROLE_PUSH_BUTTON:
 	case SPI_ROLE_CHECK_BOX:
 	case SPI_ROLE_COMBO_BOX:
-        case SPI_ROLE_SPIN_BUTTON:
+	case SPI_ROLE_SPIN_BUTTON:
 	case SPI_ROLE_RADIO_BUTTON:
-	case SPI_ROLE_PAGE_TAB:
+	/*case SPI_ROLE_PAGE_TAB:*/
 	case SPI_ROLE_TOGGLE_BUTTON:
 	case SPI_ROLE_SLIDER:
 	case SPI_ROLE_SCROLL_BAR:
-        case SPI_ROLE_LIST:
+	case SPI_ROLE_LIST:
+	case SPI_ROLE_LINK:
 		interesting = TRUE;
 		break;
 	default:
-		if (Accessible_isEditableText (accessible) ||
-		    Accessible_isHypertext (accessible) ||
+		if (gok_spy_is_editable (accessible) ||
+		    /*Accessible_isHypertext (accessible) ||*/
 		    Accessible_isTable (accessible) ||
 		    (!gok_spy_is_menu_role (role) && 
 		     (Accessible_isAction (accessible) ||
@@ -1477,6 +1549,8 @@ gok_spy_remove_node (GSList *nodes, Accessible *accessible)
 	}
 	return nodes;
 }
+
+
 /**
  * gok_spy_append_node:
  * @pNode: An existing list of AccessibleNode to append to (may be NULL).
@@ -1517,313 +1591,290 @@ GSList* gok_spy_append_node (GSList* nodes,
 	if (s)
 	  SPI_freeString (s);
 
-	/* Hypertext objects are treated differently; one node per link */
-	if (Accessible_isHypertext (pAccessible)) {
-		AccessibleHypertext *hypertext = Accessible_getHypertext (pAccessible);
-		AccessibleText *text = Accessible_getText (pAccessible);
-		gint i, nlinks = AccessibleHypertext_getNLinks (hypertext);
-		gok_log ("object contains hypertext.");
-		for (i = 0; i < nlinks; ++i) {
-			AccessibleHyperlink *link = AccessibleHypertext_getLink (hypertext, i);
-			Accessible *link_object = AccessibleHyperlink_getObject (link, 0);
-			gchar *linkname;
-			long int start, end;
-			/* TODO: handle multi-anchor links! */
-			AccessibleHyperlink_getIndexRange (link, &start, &end);
-			linkname = AccessibleText_getText (text, start, end);
-			if (linkname) {
-				/* create and append a node */
-			  /* 
-			   * FIXME: when AccessibleHyperlink objects get un-broken,
-			   * we should pass the hyperlink object here instead of the 
-			   * hypertext object, so we can just invoke the default action
-			   * on the link!  But at the moment the link objects are not
-			   * actionable, so we have to fake it.
-			   */
-				flags.data.is_link = TRUE;
-				nodes = gok_spy_add_node (nodes, pAccessible, flags, i, g_strdup (linkname));
-				SPI_freeString (linkname);
-			}
-			Accessible_unref (link_object);
-		}
-		AccessibleText_unref (text);
-		AccessibleHypertext_unref (hypertext);
-	}
-
-	/* for other node types, try to create a reasonable label string */
-	else {
+	if (pName == NULL)
+	{
 		/* look for a label */
-		if (pName == NULL)
-		{
-			gok_log("no name, so looking at relations...");
-			relations = Accessible_getRelationSet(pAccessible);
+		gok_log("no name, so looking at relations...");
+		relations = Accessible_getRelationSet(pAccessible);
 
-			if (relations != NULL)
-			while (relations[i]) {
-				type = AccessibleRelation_getRelationType(relations[i]);
-				if (type == SPI_RELATION_LABELED_BY) {
-					ntargets = AccessibleRelation_getNTargets(relations[i]);
-					for (j=0; j < ntargets; j++) {
-						targetAccessible = 
-							AccessibleRelation_getTarget(relations[i], j);
-						if (targetAccessible != NULL) {
-							s = Accessible_getName (targetAccessible);
-							
-							if (s != NULL) {
-								if ( strlen (s) != 0 )  {
-									pName = g_strdup (s);
-									SPI_freeString (s);
-									break; /* from for loop*/
-								}
-								else {
-									SPI_freeString(s);
-								}
+		if (relations != NULL)
+		while (relations[i]) {
+			type = AccessibleRelation_getRelationType(relations[i]);
+			if (type == SPI_RELATION_LABELED_BY) {
+				ntargets = AccessibleRelation_getNTargets(relations[i]);
+				for (j=0; j < ntargets; j++) {
+					targetAccessible = 
+						AccessibleRelation_getTarget(relations[i], j);
+					if (targetAccessible != NULL) {
+						s = Accessible_getName (targetAccessible);
+						
+						if (s != NULL) {
+							if ( strlen (s) != 0 )  {
+								pName = g_strdup (s);
+								SPI_freeString (s);
+								break; /* from for loop*/
+							}
+							else {
+								SPI_freeString(s);
 							}
 						}
 					}
-					break; /* from while loop */
 				}
-				i++;
+				break; /* from while loop */
 			}
+			i++;
 		}
+	}
 
-		if (pName == NULL) {
-			Accessible *parent = Accessible_getParent (pAccessible);
-			/* if there are no siblings, we can use an ancestor's name */
-			gok_log("checking ancestor names.");
-			maxloops = 7; /* ignore great great great great great great grandparent */
-			while (parent && Accessible_getChildCount (parent) == 1) 
+	if (pName == NULL) {
+		Accessible *parent = Accessible_getParent (pAccessible);
+		/* if there are no siblings, we can use an ancestor's name */
+		gok_log("checking ancestor names.");
+		maxloops = 7; /* ignore great great great great great great grandparent */
+		while (parent && Accessible_getChildCount (parent) == 1) 
+		{
+			char *tmp;
+			Accessible *tmp_parent = parent;
+			if (maxloops-- < 1) { break; }  /* guard against erroneous cyclic parent/child relations */
+			tmp = Accessible_getName (tmp_parent); 
+			gok_log ("getting parent name: parent %x, name %s, %d more loops allowed.\n", tmp_parent, tmp ? tmp : "", maxloops);
+			if (tmp != NULL) 
 			{
-				char *tmp;
-				Accessible *tmp_parent = parent;
-				if (maxloops-- < 1) { break; }  /* guard against erroneous cyclic parent/child relations */
-				tmp = Accessible_getName (tmp_parent); 
-				gok_log ("getting parent name: parent %x, name %s, %d more loops allowed.\n", tmp_parent, tmp ? tmp : "", maxloops);
-				if (tmp != NULL) 
+				if (strlen (tmp) > 0) 
 				{
-				    if (strlen (tmp) > 0) 
-				    {
-					pName = g_strdup (tmp);
-					SPI_freeString (tmp);
-					break;
-				    }
+				pName = g_strdup (tmp);
+				SPI_freeString (tmp);
+				break;
 				}
-				parent = Accessible_getParent (tmp_parent);
+			}
+			parent = Accessible_getParent (tmp_parent);
+			Accessible_unref (tmp_parent);
+		}
+	}
+	
+	if (pName == NULL) {
+		gok_log("still no name, so looking at description...");
+		/* one last try, we pull chars from description */
+		s = Accessible_getDescription (pAccessible);
+		if (s != NULL) {
+			gint len = strlen (s);
+			if ( len > 0 )  {
+				if (len > 18) 
+				{
+					gchar *tmp = NULL;
+					gok_log("shortening description");
+					tmp = g_strndup (s, 15);
+					pName = g_strconcat (tmp, "...", NULL);
+					g_free (tmp);
+				}
+				else 
+				{
+					gok_log("using description");
+					pName = g_strndup (s, len);
+				}
+			}
+			SPI_freeString (s);
+		}
+	}
+
+	if (pName == NULL) 
+	{
+		role = Accessible_getRole (pAccessible);
+	}
+
+	/* 
+	 * Role "TABLE-CELL" is used for 'generic' table children. 
+	 * These children's behavior is governed more by the fact that they are
+	 * children of a table than by their class-specific characteristics (unlike,
+	 * for instance, ROLE_CHECKBUTTON or ROLE_PUSH_BUTTON objects which may be
+	 * stored in tables).  Some table cells are composite objects, and we should
+	 * attempt to create a name string from the concatenation of their children 
+	 * when possible.  In the case of GtkTreeviews in particular, where row-selection
+	 * is the dominant user activity, this is especially appropriate.
+	 */
+	if (pName == NULL && role == SPI_ROLE_TABLE_CELL) {
+		pName = gok_spy_concatenate_child_names (pAccessible, 0);
+	}
+	else if (pName == NULL && role == SPI_ROLE_SCROLL_BAR) {
+		states = Accessible_getStateSet (pAccessible);
+		if (AccessibleStateSet_contains (states, SPI_STATE_VERTICAL))
+		{
+		/* translators: abbreviated version of "Vertical Scrollbar" */
+		pName = g_strdup (_("V Scrollbar"));
+		}
+		else
+		{
+		/* translators: abbreviated version of "Horizontal Scrollbar" */
+		pName = g_strdup (_("H Scrollbar"));
+		}
+		AccessibleStateSet_unref (states);
+	}
+	else if (pName == NULL) { /* desperate attempt to reuse singleton-child's name */
+		Accessible *tmp_parent = pAccessible;
+		gok_log("no description... checking for singleton child with name.");
+		maxloops = 3;
+		Accessible_ref (tmp_parent);
+		while (tmp_parent && Accessible_getChildCount (tmp_parent) == 1 && maxloops) 
+		{
+			Accessible *child;
+			child = Accessible_getChildAtIndex (tmp_parent, 0);
+			if (child) 
+			{
+				s = Accessible_getName (child);
+				if (s)
+				{
+					gint len = strlen (s);
+					if (len > 0) {
+						pName = g_strndup (s, len);
+						Accessible_unref (tmp_parent);
+						tmp_parent = NULL;
+					}
+					SPI_freeString (s);
+				}
+			}
+			if (tmp_parent) 
+			{
 				Accessible_unref (tmp_parent);
+				tmp_parent = child;
+			}
+			--maxloops;
+		}
+	}
+	
+	if (pName == NULL && Accessible_isTable (pAccessible)) {
+			/* Tables are so important we must include them somehow */
+		AccessibleTable *table = Accessible_getTable (pAccessible);
+		Accessible *caption = AccessibleTable_getCaption (table);
+		if (caption) 
+		{
+			char *caption_string = Accessible_getName (caption);
+			if (caption_string && strlen (caption_string)) 
+			{
+				pName = g_strndup (caption_string, 20);
+					gok_log ("Table caption: %s", caption_string);
+				SPI_freeString (caption_string);
 			}
 		}
-		
-		if (pName == NULL) {
-			gok_log("still no name, so looking at description...");
-			/* one last try, we pull chars from description */
-			s = Accessible_getDescription (pAccessible);
-			if (s != NULL) {
-				gint len = strlen (s);
-				if ( len > 0 )  {
-					if (len > 18) 
+		if (pName == NULL && AccessibleTable_getNColumns (table) == 1) 
+		{
+			char *desc = AccessibleTable_getColumnDescription (table, 0);
+			if (desc && strlen (desc)) 
+			{
+				pName = g_strndup (desc, 20);
+				SPI_freeString (desc);
+			}
+			else
+			{
+				Accessible *header;
+				header = AccessibleTable_getColumnHeader (table, 0);
+				if (header) 
+				{
+					desc = Accessible_getName (header);
+					if (desc && strlen (desc)) 
 					{
-						gchar *tmp = NULL;
-						gok_log("shortening description");
-						tmp = g_strndup (s, 15);
-						pName = g_strconcat (tmp, "...", NULL);
+							gok_log ("Table col header: %s", desc);
+						pName = g_strndup (desc, 20);
+						SPI_freeString (desc);
+					}
+					Accessible_unref (header);
+				}
+			}
+		}
+		if (pName == NULL)
+		{
+				/* translators: "table" as in row/column data structure */
+			pName = g_strdup (_("Table"));
+		}
+	}			
+
+	if (pName != NULL) {
+		gok_log("we have a name: [%s]",pName);
+		/* create a new node */
+		if (*g_strstrip (pName) == 0 || strlen (pName) == 0) 
+			g_warning ("EMPTY Node name: [%s]", pName);
+		nodes = gok_spy_add_node (nodes, pAccessible, flags, 0, pName);
+	}
+	/* if we decide to allow access to nameless menus ...
+	else if ( Accessible_getRole (pAccessible) == SPI_ROLE_MENU ) {
+		pName = g_strdup (Accessible_getRoleName(pAccessible));
+		gok_log ("Can't think of a good name so using role:[%s]",pName);
+		nodes = gok_spy_add_node (nodes, pAccessible, -1, pName);
+	}
+	*/
+	else if (Accessible_isText (pAccessible)) {
+		/* 
+		* sadly, many text objects are still nameless - yet we want 
+		* to allow the user to get to them.
+		* first tack is to get the starting text... if the object is
+		* empty, then we can still "create" a name for it, though we 
+		* can't assign it a unique meaningful name.
+		*/
+		AccessibleText *text = Accessible_getText (pAccessible);
+		gint j = 0, len;
+		gchar *word, *gs = NULL;
+
+		gok_log("still no name, but there is text...");
+
+		if (text) {
+			long int start, end = 0;
+			do {
+				gchar *tmp;
+				word = AccessibleText_getTextAtOffset (text, end, 
+									   SPI_TEXT_BOUNDARY_WORD_END,
+									   &start, &end);
+				if (word) {
+					len = strlen (word);
+				}
+				else {
+					len = 0;
+				}
+				if (len > 0) {
+					if (gs) {
+						tmp = gs;
+						gs = g_strconcat (tmp, word, NULL);
 						g_free (tmp);
 					}
-					else 
-					{
-						gok_log("using description");
-						pName = g_strndup (s, len);
-					}
-				}
-				SPI_freeString (s);
-			}
-		}
-
-		if (pName == NULL) 
-		{
-		    role = Accessible_getRole (pAccessible);
-		}
-
-		/* 
-		 * Role "TABLE-CELL" is used for 'generic' table children. 
-		 * These children's behavior is governed more by the fact that they are
-		 * children of a table than by their class-specific characteristics (unlike,
-		 * for instance, ROLE_CHECKBUTTON or ROLE_PUSH_BUTTON objects which may be
-		 * stored in tables).  Some table cells are composite objects, and we should
-		 * attempt to create a name string from the concatenation of their children 
-		 * when possible.  In the case of GtkTreeviews in particular, where row-selection
-		 * is the dominant user activity, this is especially appropriate.
-		 */
-		if (pName == NULL && role == SPI_ROLE_TABLE_CELL) {
-			pName = gok_spy_concatenate_child_names (pAccessible, 0);
-		}
-		else if (pName == NULL && role == SPI_ROLE_SCROLL_BAR) {
-		    states = Accessible_getStateSet (pAccessible);
-		    if (AccessibleStateSet_contains (states, SPI_STATE_VERTICAL))
-		    {
-			/* translators: abbreviated version of "Vertical Scrollbar" */
-			pName = g_strdup (_("V Scrollbar"));
-		    }
-		    else
-		    {
-			/* translators: abbreviated version of "Horizontal Scrollbar" */
-			pName = g_strdup (_("H Scrollbar"));
-		    }
-		    AccessibleStateSet_unref (states);
-		}
-		else if (pName == NULL) { /* desperate attempt to reuse singleton-child's name */
-			Accessible *tmp_parent = pAccessible;
-			gok_log("no description... checking for singleton child with name.");
-			maxloops = 3;
-			Accessible_ref (tmp_parent);
-			while (tmp_parent && Accessible_getChildCount (tmp_parent) == 1 && maxloops) 
-			{
-				Accessible *child;
-				child = Accessible_getChildAtIndex (tmp_parent, 0);
-				if (child) 
-				{
-					s = Accessible_getName (child);
-					if (s)
-					{
-						gint len = strlen (s);
-						if (len > 0) {
-							pName = g_strndup (s, len);
-							Accessible_unref (tmp_parent);
-							tmp_parent = NULL;
-						}
-						SPI_freeString (s);
-					}
-				}
-				if (tmp_parent) 
-				{
-					Accessible_unref (tmp_parent);
-					tmp_parent = child;
-				}
-				--maxloops;
-			}
-		}
-		if (pName == NULL && Accessible_isTable (pAccessible)) {
-		        /* Tables are so important we must include them somehow */
-			AccessibleTable *table = Accessible_getTable (pAccessible);
-			Accessible *caption = AccessibleTable_getCaption (table);
-			if (caption) 
-			{
-				char *caption_string = Accessible_getName (caption);
-				if (caption_string && strlen (caption_string)) 
-				{
-					pName = g_strndup (caption_string, 20);
-				        gok_log ("Table caption: %s", caption_string);
-					SPI_freeString (caption_string);
-				}
-			}
-			if (pName == NULL && AccessibleTable_getNColumns (table) == 1) 
-			{
-				char *desc = AccessibleTable_getColumnDescription (table, 0);
-				if (desc && strlen (desc)) 
-				{
-					pName = g_strndup (desc, 20);
-					SPI_freeString (desc);
-				}
-				else
-				{
-					Accessible *header;
-					header = AccessibleTable_getColumnHeader (table, 0);
-					if (header) 
-					{
-						desc = Accessible_getName (header);
-						if (desc && strlen (desc)) 
-						{
-						        gok_log ("Table col header: %s", desc);
-							pName = g_strndup (desc, 20);
-							SPI_freeString (desc);
-						}
-						Accessible_unref (header);
-					}
-				}
-			}
-			if (pName == NULL)
-			{
-			        /* translators: "table" as in row/column data structure */
-				pName = g_strdup (_("Table"));
-			}
-		}			
-
-		if (pName != NULL) {
-			gok_log("we have a name: [%s]",pName);
-			/* create a new node */
-			if (*g_strstrip (pName) == 0 || strlen (pName) == 0) 
-			    g_warning ("EMPTY Node name: [%s]", pName);
-			nodes = gok_spy_add_node (nodes, pAccessible, flags, 0, pName);
-		}
-		/* if we decide to allow access to nameless menus ...
-		else if ( Accessible_getRole (pAccessible) == SPI_ROLE_MENU ) {
-			pName = g_strdup (Accessible_getRoleName(pAccessible));
-			gok_log ("Can't think of a good name so using role:[%s]",pName);
-			nodes = gok_spy_add_node (nodes, pAccessible, -1, pName);
-		}
-		*/
-		else if (Accessible_isText (pAccessible)) {
-			/* 
-			* sadly, many text objects are still nameless - yet we want 
-			* to allow the user to get to them.
-			* first tack is to get the starting text... if the object is
-			* empty, then we can still "create" a name for it, though we 
-			* can't assign it a unique meaningful name.
-			*/
-			AccessibleText *text = Accessible_getText (pAccessible);
-			gint j = 0, len;
-			gchar *word, *gs = NULL;
-
-			gok_log("still no name, but there is text...");
-
-			if (text) {
-				long int start, end = 0;
-				do {
-					gchar *tmp;
-					word = AccessibleText_getTextAtOffset (text, end, 
-									       SPI_TEXT_BOUNDARY_WORD_END,
-									       &start, &end);
-					if (word) {
-						len = strlen (word);
-					}
 					else {
-						len = 0;
+						gs = g_strdup (word);
 					}
-					if (len > 0) {
-						if (gs) {
-							tmp = gs;
-							gs = g_strconcat (tmp, word, NULL);
-							g_free (tmp);
-						}
-						else {
-							gs = g_strdup (word);
-						}
-					}
-					if (word) { 
-						SPI_freeString (word);
-					}
-					++j;
-				} while ((j < 3) && len);
-				AccessibleText_unref (text);
+				}
+				if (word) { 
+					SPI_freeString (word);
+				}
+				++j;
+			} while ((j < 3) && len);
+			AccessibleText_unref (text);
+		}
+		if (gs && strlen (gs)) {
+			if (gs && g_utf8_validate (gs, -1, NULL) &&
+		  		g_utf8_strchr(gs,-1,GECKO_EMBEDDED_OBJECT_CHAR)){
+				/* We probably have a complex document, where text can contain 
+				utf-8 embed characters to denote the presence of an object 
+				within a region of text. (see mozilla a11y docs). We can safely
+				ignore this node as we capture the children denoted by the embeds
+				elsewhere. */
+				g_free(gs);
 			}
-			if (gs && strlen (gs)) {
+			else {
 				pName = g_strconcat (gs, "...", NULL);
 				g_free (gs);
 			}
-			else {
-				if (gs) {
-					g_free (gs);
-				}
-				if (Accessible_isEditableText (pAccessible)) {
-					pName = g_strdup ("Text Entry (empty)");
-				}
+		}
+		else {
+			if (gs) {
+				g_free (gs);
 			}
-			if (pName != NULL) {
-			        if (*g_strstrip (pName) == 0 || strlen (pName) == 0) 
-				    g_warning ("EMPTY Node name: [%s]", pName);
-				gok_log ("we have a name: [%s]",pName);
-				nodes = gok_spy_add_node (nodes, pAccessible, flags, 0, pName);
+			if (gok_spy_is_editable (pAccessible)) {
+				pName = g_strdup ("Text Entry (empty)");
 			}
 		}
+		if (pName != NULL) {
+			if (*g_strstrip (pName) == 0 || strlen (pName) == 0) 
+				g_warning ("EMPTY Node name: [%s]", pName);
+
+			gok_log ("we have a name: [%s]",pName);
+			nodes = gok_spy_add_node (nodes, pAccessible, flags, 0, pName);
+		}
 	}
 
 	/* use row and column info inside a table */
@@ -1947,13 +1998,14 @@ GSList* gok_spy_append_node (GSList* nodes,
 	}
 	
     /* We consider it safe enough to provide a name for nameless Selections */
-    if ((pName == NULL) && Accessible_isSelection(pAccessible)) {
+    /* removed for ff3
+	if ((pName == NULL) && Accessible_isSelection(pAccessible)) {
         gchar* rolename;
         rolename = Accessible_getRoleName (pAccessible);
         pName = g_strdup (rolename);
         nodes = gok_spy_add_node (nodes, pAccessible, flags, 0, pName);
         SPI_freeString (rolename);
-    }
+    }*/
 
    
     if ((pName == NULL) && gok_main_get_debug_nameless()) {
@@ -2576,20 +2628,19 @@ gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 				{
 				    ui_flags.data.gui = flags.data.is_ui = TRUE;
 				}
-				if (keyboard_ui_flags.data.editable_text && Accessible_isEditableText (child)) 
+				if (keyboard_ui_flags.data.editable_text && gok_spy_is_editable (child)) 
 				{
 					ui_flags.data.editable_text = TRUE;
 				}
 			}
 			if (!flags.data.is_toolbar_item) /* toolbar items already added */
 			{
-			        AccessibleNodeFlags non_context_flags = flags;
+		        AccessibleNodeFlags non_context_flags = flags;
 				non_context_flags.data.inside_html_container = FALSE;
 				if (non_context_flags.value) 
 				{
 				    _priv_ui_nodes = gok_spy_append_node (_priv_ui_nodes, child, flags);
-				    /* special case for notebook tabs and expanders, whose children should also appear */
-  				    if (role == SPI_ROLE_PAGE_TAB || role == SPI_ROLE_TOGGLE_BUTTON)
+  				    if (!gok_spy_is_menu_role(role))
   				    {
   					ui_flags.value |= gok_spy_search_component_list (child, keyboard_ui_flags, context_flags, search_depth).value;
   				    }
